#include "fileassembler.h"
#include "dirops_ew_common.h"
#include "latency_scnl.h"
#include "mseedfilename.h"
#include "ringplayer.h"
#include "transport.h"
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#define VERSION "1.5 - 2021-09-03"

#define MAX_TRACEBUF_NSAMP                                                     \
  ((int)((MAX_TRACEBUF_SIZ - sizeof(TRACE2_HEADER)) / sizeof(int)))

#define NONE "NONE"

static char *EMPTY = "";
static PARAMS params; // configuration parameters
static MSDATA latencydata;
static MSDATA *sampledataArray;
static int sampledataArraySize = 0;
static unsigned char instid, mod, type;
static char buffer[MAX_SIZE]; // Buffer for message contents
static Header *header = (Header *)buffer;
static TRACE2_HEADER *trh = (TRACE2_HEADER *)(buffer + sizeof(Header));
static int *pdata = (int *)(buffer + sizeof(Header) + sizeof(TRACE2_HEADER));
static int linenum;
static LATENCY_SCNL_INFO latency_scnl_info;

static int isNull(char *s) { return !s || strcmp(LOC_NULL_STRING, s) == 0; }

static int isWild(char *s) { return s && strcmp(s, "*") == 0; }

static int isMatch(char *s1, char *s2) {
  if (isWild(s1) || isWild(s2)) {
    return 1;
  }
  if (isNull(s1)) {
    s1 = LOC_NULL_STRING;
  }
  if (isNull(s2)) {
    s2 = LOC_NULL_STRING;
  }
  return strcmp(s1, s2) == 0;
}

/**
 * Clear the latency data.
 */
static void clearLatencyData(void) {
  clearMsdata(&latencydata);
  latencydata.debug = params.Debug;
}

/**
 * Clear the sample data.
 */
static void clearSampleData(void) {
  int i;
  MSDATA *sampledata;
  for (i = 0; i < sampledataArraySize; i++) {
    sampledata = &sampledataArray[i];
    clearMsdata(sampledata);
    sampledata->debug = params.Debug;
    sampledata->archiveFormat = params.ArchiveFormat;
    sampledata->mseeddir = params.Mseeddir;
  }
}

/**
 * Free the sample data
 */
static void freeSampleData(void) {
  int i;
  MSDATA *sampledata;
  for (i = 0; i < sampledataArraySize; i++) {
    sampledata = &sampledataArray[i];
    freeMsdata(sampledata);
  }
}

/**
 * Resize the sample data if needed
 * @param size the size
 */
static void resizeSampledata(int size) {
  if (sampledataArraySize < size) {
    sampledataArray = realloc(sampledataArray, size * sizeof(MSDATA));
    sampledataArraySize = size;
    clearSampleData();
  }
}

/**
 * Get the number of samples.
 * @param starttime the star time
 * @param endtime the end time
 * @param delta the delta
 * @return the number of samples
 */
static int getnsamp(double starttime, double endtime, double delta) {
  double diff = endtime - starttime;
  int nsamp = (diff + 0.001) / delta;
  if (diff != 1.) { // if not L1Z
    nsamp++;
  }
  return nsamp;
}

/**
 * Get data.
 * @return the data value or -999 if error (isEndData will return non-zero if
 * error)
 */
static int getData(MSDATA *msdata, double time) {
  int dataIndex;
  int data = -999;
  if (msdata->debug > 2) {
    logit("o", "getData(%f, %f, %f)\n", time, msdata->starttime,
          msdata->endtime);
  }
  if (msdata->msr == NULL) {
    if (msdata->debug > 2) {
      logit("o", "no data ready, reading first data\n");
    }
    readData(msdata);
  }
  while (!isEndData(msdata)) {
    dataIndex = round((time - msdata->starttime) / msdata->delta);
    if (dataIndex < 0) {
      if (msdata->debug > 2) {
        logit("o", "invalid data index (%d)\n", dataIndex);
        break;
      }
    }
    if (dataIndex < msdata->msr->samplecnt) {
      data = msdata->datasamples[dataIndex];
      if (msdata->debug > 2) {
        logit("o", "data[%d/%ld] (%f) (%d)\n", dataIndex,
              (long)msdata->msr->samplecnt, time, data);
      }
      return data;
    }
    // data not in buffer, read more data
    if (msdata->debug > 2) {
      logit("o", "data not in buffer, reading more data\n");
    }
    readData(msdata);
  }
  return data;
}

static double getEndTime(LATENCY_SCNL_INFO *latency_scnl) {
  double endtime;
  if (latency_scnl->starttime == latency_scnl->endtime) {
    endtime = latency_scnl->starttime + params.LatencyInterval;
  } else {
    endtime = latency_scnl->endtime;
  }
  return endtime;
}

/**
 * Process the SCNL entry.
 * @param latency_scnl the latency SCNL
 * @param entry the SCNL entry from the SCNL table
 * @param msdata the miniSEED data
 * @param outputfile the output file
 * @return 0 if sucess, positive number if error.
 */
static int processSCNLentry(LATENCY_SCNL_INFO latency_scnl, SCNL_ENTRY *entry,
                            MSDATA *msdata, char *outputfile) {
  size_t bytestowrite, byteswritten;
  int i;
  int latency;
  int nsamp;
  FILE *outfp;
  int retcode = 0;
  char *loc = latency_scnl.loc;
  char *chan = latency_scnl.chan;
  double time;
  double starttime = latency_scnl.starttime;
  double endtime = getEndTime(&latency_scnl);

  latency = latency_scnl.latency;

  // if entry found
  if (entry) {
    loc = entry->loc;
    chan = entry->chan;
  }

  if (isNull(loc)) {
    loc = EMPTY;
  }

  MseedFileName(msdata->mseedfile, msdata->mseeddir, starttime,
                latency_scnl.sta, chan, latency_scnl.net, loc,
                msdata->archiveFormat);
  if (*msdata->mseedfile == '\0') {
    logit("e", "could not find miniSEED file %f (%s.%s.%s.%s)\n", starttime,
          latency_scnl.sta, latency_scnl.net, loc, chan);
    return 1;
  }

  if (params.Debug > 2) {
    logit("o", "reading initial data for line %d '%s'\n", linenum, buffer);
  }
  readData(msdata);
  if (isEndData(msdata)) {
    logit("e", "could not read data %f (%s.%s.%s.%s)\n", starttime,
          latency_scnl.sta, latency_scnl.net, loc, chan);
    return 2;
  }
  nsamp = getnsamp(starttime, endtime, msdata->delta);
  if (nsamp > MAX_TRACEBUF_NSAMP) {
    logit("e", "nsamp (%d) is larger than max (%d)\n", nsamp,
          MAX_TRACEBUF_NSAMP);
    return 3;
  }

  if (params.Debug > 2) {
    logit("o", "mseedfile (%f, %f, %f, %d)=%s\n", starttime, endtime,
          msdata->msr->samprate, nsamp, msdata->mseedfile);
  }

  memset(buffer, 0, MAX_SIZE);
  header->instid = instid;
  header->mod = mod;
  header->type = type;
  time = endtime;
  time += (latency / 1000.); // add latency converted from ms to seconds
  header->sec = (int64_t)time;
  header->usec = (int64_t)((time - header->sec) * 1000000.);
  // length of TBUF2 data (does not include the Header)
  header->length = sizeof(TRACE2_HEADER) + nsamp * 4;
  bytestowrite = sizeof(Header) + header->length;

  // set trace header fields
  trh->version[0] = TRACE2_VERSION0;
  trh->version[1] = TRACE2_VERSION1;
  trh->pinno = params.Pinno;
  trh->nsamp = nsamp;
  trh->starttime = starttime;
  trh->endtime = endtime;
  trh->samprate = msdata->msr->samprate;
  strcpy(trh->sta, latency_scnl.sta);
  strcpy(trh->chan, latency_scnl.chan);
  strcpy(trh->net, latency_scnl.net);
  strcpy(trh->loc, latency_scnl.loc);
#ifdef _INTEL
  strcpy(trh->datatype, "i4");
#endif
#ifdef _SPARC
  strcpy(trh->datatype, "s4");
#endif
  // trh->quality;
  // trh->pad;

  // get the data from the miniSEED file
  for (i = 0; i < nsamp; i++) {
    time = starttime + i * msdata->delta;
    pdata[i] = getData(msdata, time);
    if (isEndData(msdata)) {
      logit("e", "could not get data (%d/%d, %f)\n", i, nsamp, time);
      retcode |= 1;
      break;
    }
  }
  if (retcode != 0) {
    return retcode;
  }
  outfp = OpenFile(outputfile, "ab");
  if (outfp == NULL) {
    logit("e", "could not open file (%s): %s\n", outputfile, strerror(errno));
    return 4;
  }
  // write to the output file
  byteswritten = fwrite(buffer, 1, bytestowrite, outfp);
  if (bytestowrite != byteswritten) {
    logit("e", "erring writing file (%s): %s\n", outputfile, strerror(errno));
  }
  fclose(outfp);
  return retcode;
}

/**
 * Process the SCNL.
 * @param latency_scnl the latency SCNL
 * @param outputfile the output file
 * @return 0 if sucess or ignored, positive number if error.
 */
static int processSCNL(LATENCY_SCNL_INFO latency_scnl, char *outputfile) {
  int i;
  int retcode = 0;
  SCNL_ENTRY *entry = NULL, *entry_match = NULL;
  double starttime = latency_scnl.starttime;
  double endtime = getEndTime(&latency_scnl);

  if (params.StartTime) { // if start time specified
    // if this data is before the start time
    if (starttime < params.StartTime) {
      if (params.Debug > 1) {
        logit("o", "processSCNL ignore (%s.%s.%s.%s) starttime=%f\n",
              latency_scnl.sta, latency_scnl.chan, latency_scnl.net,
              latency_scnl.loc, starttime);
      }
      return 0; // ignore the data
    }
  }

  if (params.EndTime) { // if end time specified
    // if this data is after the end time
    if (endtime > params.EndTime) {
      if (params.Debug > 1) {
        logit("o", "processSCNL ignore (%s.%s.%s.%s) endtime=%f\n",
              latency_scnl.sta, latency_scnl.chan, latency_scnl.net,
              latency_scnl.loc, endtime);
      }
      return 0; // ignore the data
    }
  }

  if (strcmp(params.Mseeddir, NONE) == 0) {
    write_latency(&latency_scnl, stdout, params.LatencyOutputType);
    return 0;
  }
  if (params.Debug > 0) {
    write_latency(&latency_scnl, stdout, params.LatencyOutputType);
  }

  if (params.scnl_table.nSCNL) {
    for (i = 0; i < params.scnl_table.nSCNL; i++) {
      entry = params.scnl_table.scnl_entries + i;
      // if SCNL matches
      if (isMatch(latency_scnl.net, entry->net) &&
          isMatch(latency_scnl.sta, entry->sta) &&
          (params.MatchAllChannelCodes ||
           isMatch(latency_scnl.chan, entry->chan)) &&
          (params.MatchAllLocationCodes ||
           isMatch(latency_scnl.loc, entry->loc))) {
        entry_match = entry;
        if (params.Debug > 0) {
          logit("o", "found match for SCNL (%s.%s.%s.%s)\n", entry->sta,
                entry->chan, entry->net, entry->loc);
        }
        resizeSampledata(params.scnl_table.nSCNL);
        retcode = processSCNLentry(latency_scnl, entry, &sampledataArray[i],
                                   outputfile);
        if (retcode) {
          break;
        }
      }
    }
    if (entry_match == NULL) {
      if (params.Debug > 0) {
        logit("o", "could not find match for SCNL (%s.%s.%s.%s)\n",
              latency_scnl.sta, latency_scnl.chan, latency_scnl.net,
              latency_scnl.loc);
      }
    }
  } else {
    resizeSampledata(1);
    retcode =
        processSCNLentry(latency_scnl, entry, &sampledataArray[0], outputfile);
  }
  return retcode;
}

/**
 * Process the latency from L1Z
 * @param latencyfile the latency file
 * @param outputfile the output file
 * @return 0 if sucess or ignored, positive number if error.
 */
static int processLatencyL1Z(char *latencyfile, char *outputfile) {
  int retcode = 0;

  clearLatencyData();
  strncpy(latencydata.mseedfile, latencyfile, MSEEDFILE_SZ);
  while (!isEndData(&latencydata)) {
    readData(&latencydata);
    if (latencydata.debug > 0) {
      logit("o", "processLatencyMsFile(%s)\n", latencydata.mseedfile);
    }
  }
  return retcode;
}

/**
 * Process the latency from latency SCNL file
 * @param latencyfile the latency file
 * @param outputfile the output file
 * @return 0 if success, -1 if EOF, positive number if error.
 */
static int processLatencyScnlFileBin(char *latencyfile, char *outputfile) {
  int retcode = 0;
  FILE *infp;

  clearSampleData();

  // read the input file
  infp = fopen(latencyfile, "rb");
  if (infp == NULL) {
    logit("e", "could not open file (%s): %s\n", latencyfile, strerror(errno));
    retcode = errno;
    return retcode;
  }

  for (linenum = 1;; linenum++) {
    freeSampleData(); // free sample data if needed
    retcode = read_latency_bin(&latency_scnl_info, infp);
    if (retcode) {
      if (retcode < 0) { // if EOF
        break;
      }
      logit("e", "could not read latency packet %d (%s)\n", linenum,
            strerror(retcode));
      break;
    }
    retcode = processSCNL(latency_scnl_info, outputfile);
  } // end for line
  fclose(infp);
  freeSampleData(); // free sample data if needed
  return retcode;
}

/**
 * Process the latency from latency SCNL file
 * @param latencyfile the latency file
 * @param outputfile the output file
 * @return 0 if success, -1 if EOF, positive number if error.
 */
static int processLatencyScnlFileText(char *latencyfile, char *outputfile) {
  int retcode = 0;
  FILE *infp;

  clearSampleData();

  // read the input file
  infp = fopen(latencyfile, "r");
  if (infp == NULL) {
    if (errno) {
      retcode = errno;
      logit("e", "could not open file (%s): %s\n", latencyfile,
            strerror(errno));
    } else {
      retcode = -1;
    }
    return retcode;
  }

  for (linenum = 1;; linenum++) {
    freeSampleData(); // free sample data if needed
    retcode = read_latency_text(&latency_scnl_info, infp);
    if (retcode) {
      if (retcode < 0) { // if EOF
        break;
      }
      logit("e", "could not parse latency on line %d\n", linenum);
      break;
    }
    retcode = processSCNL(latency_scnl_info, outputfile);
  } // end for line
  fclose(infp);
  freeSampleData(); // free sample data if needed
  return retcode;
}

/**
 * Process the latency file
 * @param latencyfile the latency file
 * @param outputfile the output file
 * @return 0 if success, -1 if EOF, positive number if error.
 */
static int processFile(char *latencyfile, char *outputfile) {
  int retcode;
  if (strendswith(latencyfile, ".ms") ||
      strendswith(latencyfile, ".miniseed")) {
    retcode = processLatencyL1Z(latencyfile, outputfile);
  } else if (params.Binary) {
    retcode = processLatencyScnlFileBin(latencyfile, outputfile);
  } else {
    retcode = processLatencyScnlFileText(latencyfile, outputfile);
  }
  return retcode;
}

void printUsage(char *progname) {
  fprintf(stderr,
          "Usage:   %s <configfile> [--option=value]\n"
          "or to dump the file contents to stdout:\n"
          "         %s -d <inputfile> [option value]\n",
          progname, progname);
  fprintf(stderr, "\nWhere option may be:\n");
  printConfigHelp(stderr);
  fprintf(stderr, "Version: %s\n", VERSION);
}

int main(int argc, char *argv[]) {
  int isLatencyFile;
  int len = 0;
  int retcode = 0;
  const char *prog;

  if (argc < 2) {
    printUsage(argv[0]);
    return -1;
  }
  initParams(&params);
  if (strcmp("-d", argv[1]) == 0) {
    if (argc <= 2) {
      printUsage(argv[0]);
      return -1;
    }
    prog = argv[0];
    logit_init(prog, 0, 256, params.LogFile);
    strncpy(params.Latencyfile, argv[2], sizeof(params.Latencyfile));
    strcpy(params.Outputfile, NONE);
    strcpy(params.Mseeddir, NONE);
    if (argc >= 3) {
      retcode = config(argc, argv, &params, 3);
      if (retcode) {
        printUsage(argv[0]);
        return retcode;
      }
    }
  } else {
    prog = argv[1];
    logit_init(prog, 0, 256, params.LogFile);
    if (GetLocalInst(&instid) != 0) {
      logit("e", "%s: error getting local installation id; exiting!\n",
            argv[0]);
      return -1;
    }
    if (GetModId("MOD_WILDCARD", &mod) != 0) {
      logit("e", "%s: Invalid module name <MOD_WILDCARD>; exiting!\n", argv[0]);
      return -1;
    }
    if (GetType("TYPE_TRACEBUF2", &type) != 0) {
      logit("e",
            "%s: Message of type <TYPE_TRACEBUF2> not found in earthworm.d or "
            "earthworm_global.d; exiting!\n",
            argv[0]);
      return -1;
    }
    retcode = config(argc, argv, &params, 0);
    if (retcode) {
      printUsage(argv[0]);
      return retcode;
    }
  }
  if (params.LogFile) {
    logit_init(prog, 0, 256, params.LogFile);
  }

  if (!*params.Latencyfile || !*params.Outputfile || !*params.Mseeddir) {
    logit("e", "fileassembler: ERROR, no ");
    if (!*params.Latencyfile)
      logit("e", "<Latencyfile> ");
    if (!*params.Outputfile)
      logit("e", "<Outputfile> ");
    if (!*params.Mseeddir)
      logit("e", "<Mseeddir> ");
    logit("e", "command(s) in <%s>; exiting!\n", argv[1]);
    return -2;
  }

  // if no output file extension and input and output file are the same
  if (!*params.OutputfileExt &&
      strcmp(params.Latencyfile, params.Outputfile) == 0) {
    logit("e", "<Latencyfile> cannot be the same as <Outputfile> unless "
               "<OutputfileExt> is specified\n");
    return -3;
  }

  isLatencyFile = 0;
  if (*params.LatencyfileExt) { // if latency file extension specified
    // if latency file ends with the latency file extension
    if (strendswith(params.Latencyfile, params.LatencyfileExt)) {
      isLatencyFile = 1;
    }
  }
  if (!isLatencyFile &&
      OpenDir(params.Latencyfile) == 0) { // if Latencyfile is a directory
    char fname[256];
    char latencyfile[2048];
    char outputfile[2048];
    while (GetNextFileName(fname) == 0) {
      if (*params.LatencyfileExt) { // if latency file extension specified
        int fnamelen = strlen(fname);
        len = strlen(params.LatencyfileExt);
        // if file extension does not match
        if (fnamelen <= len ||
            strcmp(params.LatencyfileExt, fname + fnamelen - len) != 0) {
          if (params.Debug > 0) {
            logit("o", "processFile skipping fname=%s\n", fname);
          }
          continue; // ignore the file
        }
      }
      sprintf(latencyfile, "%s/%s", params.Latencyfile, fname);
      sprintf(outputfile, "%s/%s", params.Outputfile, fname);
      if (*params.OutputfileExt) { // if output file extension specified
        if (len > 0) {
          // remove the file extension
          outputfile[strlen(outputfile) - len] = '\0';
        }
        len = strlen(outputfile);
        strncpy(outputfile + len, params.OutputfileExt,
                sizeof(outputfile) - len);
      }
      if (params.Debug > 0) {
        logit("o", "calling processFile(%s, %s)\n", latencyfile, outputfile);
      }
      retcode = processFile(latencyfile, outputfile);
    }
  } else if (*params.LatencyfileExt && !isLatencyFile) {
    logit("e", "<Latencyfile> (%s) does not match <LatencyfileExt> (%s)\n",
          params.Latencyfile, params.LatencyfileExt);
    return -3;
  } else {                       // if Latencyfile is not a directory
    if (*params.OutputfileExt) { // if output file extension specified
      // add the output file extension
      len = strlen(params.Outputfile);
      strncpy(params.Outputfile + len, params.OutputfileExt,
              sizeof(params.Outputfile) - len);
    }
    if (params.Debug > 0) {
      logit("o", "processFile(%s, %s)\n", params.Latencyfile,
            params.Outputfile);
    }
    retcode = processFile(params.Latencyfile, params.Outputfile);
  }
  return retcode;
}
