
/*****************************************************************
 *      Name: fileplayer (was file2ring)                         *
 *                                                                *
 *  Author: Andrew Good                                           *
 *                                                                *
 *  Description:                                                  *
 *                                                                *
 *  This program takes as input a file of the type created        *
 *  by "ring2file", which consists of messages with headers       *
 *  that contain the message logo and original time of receipt.   *
 *  Message content is the exact sequence of bytes that was read  *
 *  from the ring by "ring2file".  These messages are placed on   *
 *  an output Earthworm ring with a cadence determined from       *
 *  analysis of the receipt times in the headers, to simulate     *
 *  the behavior of input ring that was monitored by "ring2file". *
 *                                                                *
 ******************************************************************/

/*
 *
 * Copyright (c) 2017 California Institute of Technology.
 *
 * All rights reserved.
 *
 * This program is distributed WITHOUT ANY WARRANTY whatsoever.
 *
 * Do not redistribute this program without written permission.
 *
 */

#include "ringplayer.h"
#include <earthworm.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <time_ew.h>
#include <trace_buf.h>
#include <transport.h>
#include <unistd.h>
#include "kom.h"
#include "lmplatform.h"

#define VERSION "1.1 - 2021-08-20"

static const char *MYNAME = "fileplayer";

static unsigned char TypeTraceBuf2;
static unsigned char TypePickSCNL;
static unsigned char TypeQuake2K;
static unsigned char TypeArc;
static unsigned char TypeCarlStaTrig;
static unsigned char TypeCarlSubTrig;
static unsigned char TypeHeartbeat;
static unsigned char TypeError;
static int ScreenMsg = 1;   // 1 to write info to screen, 0 to omit.
static int debug = 1;       // Toggle debug messages
static int fixedInterval = 0;  // Whether to play back at original cadence or fixed rate
static int dump = 0;
static int makeCurrent = 0; // Offset timestamps to simulate "new" data
static double interval = 100.0; // Inter-message interval (fixed rate mode)
static double playSpeed = 1.0;  // Playback rate (original cadence mode)
static time_t startTime = 0;
static time_t endTime = 0;
static char fileName[1024];   // Input file name
static char outRing[1024];    // Output ring name
static int use_window = 0; // Whether or not to play back only a specific time window
static int stpoutput = 0;  // Whether or not to include STP test output

void lookup(void);
void usage(void);

// Read the configuratin file, return zero if success, non-zero otherwise
static int read_config(char *configFile) {
  char *com;
  if (k_open(configFile) == 0) {
    fprintf(stderr, "Error: Unable to open configuration file \"%s\"\n", configFile);
    fflush(stderr);
    return 1;
  }
  while (k_rd())
  {
    com = k_str(); /* Get the first token from line */

    // Ignore blank lines
    if(!com)
      continue;

    // Ignore comments
    if(com[0] == '#')
      continue;

    if (strcasecmp(com, "debug") == 0) {
      debug = atoi(k_str());
      if(debug) {
        printf("debug=%d\n", debug);
      }
    } else if (strcasecmp(com, "makeCurrent") == 0) {
      makeCurrent = atoi(k_str());
      if(debug) {
        printf("makeCurrent=%d\n", makeCurrent);
      }
    } else if (strcasecmp(com, "fixedInterval") == 0) {
      fixedInterval = 1;
      interval = atof(k_str());
      if(debug) {
        printf("fixedInterval=%f\n", interval);
      }
    } else if (strcasecmp(com, "playbackRate") == 0) {
      playSpeed = atof(k_str());
      fixedInterval = 0;
      if(debug) {
        printf("playbackRate=%f\n", playSpeed);
      }
    } else if (strcasecmp(com, "infileName") == 0) {
      strcpy(fileName, k_str());
      if(debug) {
        printf("infileName=%s\n", fileName);
      }
    } else if (strcasecmp(com, "ringName") == 0) {
      strcpy(outRing, k_str());
      if(debug) {
        printf("ringName=%s\n", outRing);
      }
    } else if (strcasecmp(com, "startTime") == 0) {
      startTime = atol(k_str());
      if(debug) {
        printf("startTime=%ld\n", startTime);
      }
    } else if (strcasecmp(com, "endTime") == 0) {
      endTime = atol(k_str());
      if(debug) {
        printf("endTime=%ld\n", endTime);
      }
    } else if (strcasecmp(com, "stp") == 0) {
      stpoutput = 1;
      if(debug) {
        printf("stpoutput=%d\n", stpoutput);
      }
    } else {
       printf("Error: unexpected command (%s)\n", com);
       return -1;
    }
  }
  return 0;
}

// Process the configuration options, return zero if success, non-zero otherwise
static int process_config(int argc, char *argv[]) {
  int argi, i, count = 0;
  // Check command line statements
  if (argc < 2) {
    printf("Error: argc < 2\n");
    return -1;
  }
  
  // Check if configuration file was specified
  argi = 1;
  i = strlen(argv[argi]) - 2;
  if (i > 0 && strcmp(argv[argi]+i, ".d")==0) {
    i = read_config(argv[argi]);
    if (i != 0) {
      return i;
    }
    argi++;
  }
  // Check other arguments
  for (i = argi; i < argc; i++) {
    if (strcasecmp(argv[i], "--debug") == 0) {
      if (++i >= argc) {
         printf("Error: debug value must be specified\n");
         return -1;
      }
      debug = atoi(argv[i]);
      if(debug > 1) {
        printf("debug=%d\n", debug);
      }
    } else if (strcmp(argv[i], "--stp") == 0) {
      stpoutput = 1;
      if(debug) {
        printf("stpoutput=%d\n", stpoutput);
      }
    } else if (strcmp(argv[i], "-d") == 0) {
      dump = 1;
      if(debug > 1) {
        printf("dump=%d\n", dump);
      }
    } else if (strcmp(argv[i], "-c") == 0) {
      makeCurrent = 1;
      if(debug > 1) {
        printf("makeCurrent=%d\n", makeCurrent);
      }
    } else if (strcmp(argv[i], "-f") == 0) {
      if (++i >= argc) {
         printf("Error: fixedInterval must be specified\n");
         return -1;
      }
      fixedInterval = 1;
      interval = atof(argv[i]);
      if(debug > 1) {
        printf("fixedInterval=%f\n", interval);
      }
    } else if (strcmp(argv[i], "-r") == 0) {
      if (++i >= argc) {
         printf("Error: playbackRate must be specified\n");
         return -1;
      }
      playSpeed = atof(argv[i]);
      if(debug > 1) {
        printf("playbackRate=%f\n", playSpeed);
      }
    } else {
      switch (++count) {
      case 1: // infileName
         strcpy(fileName, argv[i]);
         if(debug > 1) {
           printf("infileName=%s\n", fileName);
         }
         if (dump) {
           ++count; // skip ringName
         }
         break;
      case 2: // ringName
         strcpy(outRing, argv[i]);
         if(debug > 1) {
           printf("ringName=%s\n", outRing);
         }
         break;
      case 3: // startTime
         use_window = 1;
         startTime = (time_t)strtol(argv[i], NULL, 10);
         if(debug > 1) {
           printf("startTime=%ld\n", startTime);
         }
         break;
      case 4: // endTime
         endTime = (time_t)strtol(argv[i], NULL, 10);
         if(debug > 1) {
           printf("endTime=%ld\n", endTime);
         }
         break;
      default:
         printf("Error: unexpected argument (%s)\n", argv[i]);
         return -1;
      }
    }
  }
  // Check for missing or invalid arguments
  if (*fileName == '\0') {
     printf("infileName must be specified\n");
     return -1;
  }
  if (!dump && *outRing == '\0') {
     printf("ringName must be specified\n");
     return -1;
  }
  if (use_window && endTime < startTime) {
    printf("endTime cannot be less than startTime\n");
    return -1;
  }
  return 0;
}

// Main method
int main(int argc, char *argv[]) {

  int first = 1; // Whether this is the first record in the input file
  int firstout = 0;  // Whether the first record information has been output yet
  int fd;             // Input file descriptor
  int nread = 0;      // Amount read into buffer in previous "read" command
  int i;

  long outputKey; // Key to the transport ring we write to

  // Double-precision times for printouts; less accurate than
  // storing as second/microsecond pairs.

  double firstTime = 0.0; // Time of receipt of first record
  double prevTime = 0.0;  // Time of receipt of previous record
  double nextTime = 0.0;  // Time of receipt of subsequent record
  double secOffset = 0.0; // Seconds to offset for simulating "new" data
  double sleepTime = 0.0; // Variable inter-record sleeptime for playback

  double CurrentTime;      // Current system time used in packet timing
  double lastdot = 0.;     // Time last dot was written to screen

  double starttime_orig = 0.0;
  double endtime_orig = 0.0;

  // Microsecond-precision versions of time quanitites described above.
  // These are only really necessary when timing playback of wave ring
  // records; other types do not require this level of precision.

  int64_t firstTime_sec = 0;
  int64_t firstTime_usec = 0;
  int64_t prevTime_sec = 0;
  int64_t prevTime_usec = 0;
  int64_t nextTime_sec = 0;
  int64_t nextTime_usec = 0;
  int64_t offset_sec = 0;
  int64_t offset_usec = 0;
  int64_t sleepTime_sec = 0;
  int64_t sleepTime_usec = 0;

  // double drift       = 0;

  struct timeval tp;   // Parameter for calling gettimeofday()
  struct timezone tzp; // Parameter for calling gettimeofday()


  char buffer[MAX_SIZE]; // Buffer for message contents


  // long       outputLength;
  // char  outmsg[MAX_SIZE];

  SHM_INFO outRegion;  // Shared memory information about output ring
  MSG_LOGO outputLogo; // Message description ("transport.h")
  TRACE2_HEADER *trh;

  // Earthworm 7.9 is set up to support up to 256 different message types.
  int activeTypes[256];

  for (i = 0; i < 256; ++i) {
    activeTypes[i] = 0;
  }

  activeTypes[TYPE_WILDCARD] = 1;
  activeTypes[TYPE_ADBUF] = 1;
  activeTypes[TYPE_ERROR] = 0;
  activeTypes[TYPE_HEARTBEAT] = 0;
  activeTypes[TYPE_TRACE2_COMP_UA] = 1;
  activeTypes[TYPE_NANOBUF] = 1;
  activeTypes[TYPE_ACK] = 1;
  activeTypes[TYPE_PICK_SCNL] = 1;
  activeTypes[TYPE_CODA_SCNL] = 1;
  activeTypes[TYPE_PICK2K] = 1;
  activeTypes[TYPE_CODA2K] = 1;
  activeTypes[TYPE_PICK2] = 1;
  activeTypes[TYPE_CODA2] = 1;
  activeTypes[TYPE_HYP2000ARC] = 1;
  activeTypes[TYPE_H71SUM2K] = 1;
  //    activeTypes[TYPE_EVENT_SCNL]                    = 0;
  activeTypes[TYPE_HINVARC] = 1;
  activeTypes[TYPE_H71SUM] = 1;
  activeTypes[TYPE_TRACEBUF2] = 1;
  activeTypes[TYPE_TRACEBUF] = 1;
  activeTypes[TYPE_LPTRIG] = 1;
  activeTypes[TYPE_CUBIC] = 1;
  activeTypes[TYPE_CARLSTATRIG] = 1;
  activeTypes[TYPE_TRIGLIST] = 1;
  activeTypes[TYPE_TRIGLIST2K] = 1;
  activeTypes[TYPE_TRACE_COMP_UA] = 1;
  activeTypes[TYPE_STRONGMOTION] = 1;
  activeTypes[TYPE_MAGNITUDE] = 1;
  activeTypes[TYPE_STRONGMOTIONII] = 1;
  activeTypes[TYPE_LOC_GLOBAL] = 1;
  activeTypes[TYPE_LPTRIG_SCNL] = 1;
  activeTypes[TYPE_CARLSTATRIG_SCNL] = 1;
  activeTypes[TYPE_TRIGLIST_SCNL] = 1;
  activeTypes[TYPE_MSEED] = 1;
  activeTypes[TYPE_ACTIVATE_MODULE] = 1;
  activeTypes[TYPE_SPECTRA] = 1;
  activeTypes[TYPE_QUAKE2K] = 1;
  activeTypes[TYPE_LINK] = 1;
  activeTypes[TYPE_EVENT2K] = 1;
  activeTypes[TYPE_PAGE] = 1;
  activeTypes[TYPE_KILL] = 1;
  activeTypes[TYPE_DSTDRINK] = 1;
  activeTypes[TYPE_RESTART] = 1;
  activeTypes[TYPE_REQSTATUS] = 1;
  activeTypes[TYPE_STATUS] = 1;
  activeTypes[TYPE_EQDELETE] = 1;
  activeTypes[TYPE_EVENT_SCNL] = 1;
  activeTypes[TYPE_RECONFIG] = 1;
  activeTypes[TYPE_STOP] = 1;
  activeTypes[TYPE_CANCELEVENT] = 1;
  activeTypes[TYPE_EVENT_SIGNAL] = 1;
  activeTypes[TYPE_EC_EVENT] = 1;
  activeTypes[TYPE_EC_CANCEL] = 1;
  activeTypes[TYPE_TRIG_SIGNAL] = 1;
  activeTypes[TYPE_TRIMAG_MAG] = 1;
  activeTypes[TYPE_K2INFO_PACKET] = 1;

  Header header;

  i = process_config(argc, argv);
  if (i) {
    usage();
    return i;
  }
  if (stpoutput) {
    if (use_window) {
      fprintf(stderr, ">> startTime = %ld\n", startTime);
      fprintf(stderr, ">> endTime   = %ld\n", endTime);
    }
    fprintf(stderr, "use_window = %d\n", use_window);
    fprintf(stderr, "fixedInterval = %d\n", fixedInterval);
    if(fixedInterval) {
      fprintf(stderr, ">> interval = %f\n", interval);
    }
    fflush(stderr);
  }
  if (debug > 1) {
    if (use_window) {
      printf("Time window: startTime=[%ld], endTime=[%ld]\n", startTime,
             endTime);
    } else {
      printf("No time window, play back entire file.\n");
    }
  }

  // Look up logos
  lookup();

  if (!dump) {
    // Attach to output ring
    if ((outputKey = GetKey(outRing)) == -1) {
      fprintf(stderr, "Error: Invalid RingName; exiting...\n");
      fflush(stderr);
      return -1;
    }

    tport_attach(&outRegion, outputKey);

    if (debug > 1) {
      fprintf(stderr, "Attached to output ring.\n");
      fflush(stderr);
    }

    // Get reference time
    gettimeofday(&tp, &tzp);
    // hrtime_ew( &d_now );
  }

  // Open input file
  fd = open(fileName, O_RDONLY);
  lseek(fd, 0, SEEK_SET);

  if (stpoutput) {
    printf("\nDEBUG: Begin playback.\n");
    fflush(stdout);
  }

  // Main loop
  while (1) {
    // Read (constant-size) header for message
    nread = read(fd, &header, sizeof(header));

    if (nread <= 0)
      break;

    if (debug > 1) {
      printf("DEBUG: nread = [%d]\n", nread);
      printf("DEBUG: header.length = [%d]\n", (int)header.length);
      fflush(stdout);
    }

    // Read message to place on output ring
    nread = read(fd, buffer, header.length);

    if (nread <= 0)
      break;

    if (debug > 1) {
      printf("DEBUG: nread = [%d]\n", nread);
      fflush(stdout);
    }

    // Append null character in case previous message was longer
    buffer[header.length] = '\0';

    // TODO:
    // To test whether or not the same inputs to AQMS result in the same
    // outputs, it may be useful to simulate "new" events that are otherwise
    // identical to "old" events, to see whether or not the information that
    // makes it into the database is consistent with the results for those old
    // events.  This would entail modifying timestamps for different types of
    // records with an offset (which itself would entail recognizing the record
    // type, which this program currently does not do for any records other than
    // TRACE_BUF2).

    if (use_window) {
      if (header.sec < startTime)
        continue;

      if (header.sec > endTime)
        break;
    }

    if (dump) {
      printf("sec           = [%ld]\n", (long)header.sec);
      printf("usec          = [%ld]\n", (long)header.usec);
      printf("length        = [%ld]\n", (long)header.length);
      if (debug > 1) {
        printf("type          = [%d]\n", header.type);
        printf("mod           = [%d]\n", header.mod);
        printf("instid        = [%d]\n", header.instid);
      }
      if (header.type == TypeTraceBuf2) { // Wave records
         int *data;
         trh = (TRACE2_HEADER *)buffer;
         printf("starttime     = [%f]\n", trh->starttime);
         printf("endtime       = [%f]\n", trh->endtime);
         printf("samprate      = [%f]\n", trh->samprate);
         printf("S.C.N.L       = [%s.%s.%s.%s]\n",
           trh->sta, trh->chan, trh->net, trh->loc);
         printf("nsamp         = [%d]\n", trh->nsamp);
         if (debug > 1) {
           printf("datatype      = [%s]\n", trh->datatype);
         }
         data = (int *)(buffer + sizeof(TRACE2_HEADER));
         if (debug > 2) {
           printf("data          =\n");
           for (i=0; i<trh->nsamp; ++i) {
             printf("%6d ", data[i]);
             if(i%10==9) printf("\n");
           }
           printf("\n");
         }
      } else {
        printf(" --> Message: %s", (char *)buffer);
      }
      printf("\n");
      fflush(stdout);
      continue;
    }

    if (first) {
      if (debug > 1 || stpoutput) {
        printf("DEBUG: First Record -->\n");
        fflush(stdout);
      }

      firstTime = (double)header.sec * 1000000. + (double)header.usec / 1.;
      secOffset =
          (tp.tv_sec * 1. + tp.tv_usec / 1000000.) - firstTime / 1000000.;

      firstTime_sec = header.sec;
      firstTime_usec = header.usec;

      offset_sec = tp.tv_sec - firstTime_sec;
      if (offset_sec < 0) {
        printf("ERROR: firstTime > currentTime\n");
        return -1;
      }
      offset_usec = tp.tv_usec - firstTime_usec;
      if (offset_usec < 0) {
        offset_sec -= 1;
        offset_usec += 1000000;
      }

      if (debug > 1 || stpoutput) {
        printf("DEBUG: tp.tv_sec      = [%d]\n", (int)tp.tv_sec);
        printf("DEBUG: tp_tv_usec     = [%d]\n", (int)tp.tv_usec);
        printf("DEBUG: firstTime_sec  = [%d]\n", (int)firstTime_sec);
        printf("DEBUG: firstTime_usec = [%d]\n", (int)firstTime_usec);
        printf("DEBUG: offset_sec     = [%d]\n", (int)offset_sec);
        printf("DEBUG: offset_usec    = [%d]\n", (int)offset_usec);
        fflush(stdout);
      }

      prevTime = firstTime;
      nextTime = firstTime;

      prevTime_sec = firstTime_sec;
      prevTime_usec = firstTime_usec;
      nextTime_sec = firstTime_sec;
      nextTime_usec = firstTime_usec;

      first = 0;
    }

    if (debug > 1) {
      printf("startTime=[%ld], endTime=[%ld], header.sec=[%lld]\n", startTime,
             endTime, (long long int) header.sec);
      fflush(stdout);
    }

    if (debug > 1) {
      printf("[%d]", (int)header.sec);
      printf("[%d]", (int)header.usec);
      printf("[%d]", (int)header.length);
      printf("[%d]", (int)header.type);
      printf("[%d]", (int)header.mod);
      printf("[%d]", (int)header.instid);
      printf(" --> Message: %s", (char *)buffer);
      printf("\n");
      fflush(stdout);
    }

    // Construct output message logo
    outputLogo.type = header.type;
    outputLogo.mod = header.mod;
    outputLogo.instid = header.instid;

    nextTime = (double)header.sec * 1000000. + (double)header.usec / 1.;
    nextTime_sec = header.sec;
    nextTime_usec = header.usec;

    // The following calculation assumes that messages are in time order; since
    // the timestamps are system clock times when messages were received, as
    // long as the system clock is operating properly, this will always be the
    // case.

    if (outputLogo.type == TypeTraceBuf2) // Wave records
    {
      trh = (TRACE2_HEADER *)buffer;

      starttime_orig = trh->starttime;
      endtime_orig = trh->endtime;
    }

    useconds_t sleep_usec = 0;

    if (fixedInterval) {
      // Since outputting records at fixed intervals in no way replicates
      // anything that happens in real life, this option is included primarily
      // for debugging purposes.

      sleep_ew(interval * 1000.);
    } else {
      gettimeofday(&tp, &tzp);
      double unow = tp.tv_sec * 1000000. + tp.tv_usec / 1.;

      int64_t now_sec = tp.tv_sec;
      int64_t now_usec = tp.tv_usec;

      if (debug > 1) {
        printf("now_sec=[%d] now_usec=[%d]\n", (int)now_sec, (int)now_usec);
        fflush(stdout);
      }

      // if(playSpeed == 1.0)
      // {
      sleepTime = (nextTime + 1000000. * secOffset) - unow;

      int64_t updatedTime_sec = 0;
      int64_t updatedTime_usec = nextTime_usec + offset_usec;
      int64_t carry_sec = (int64_t)(updatedTime_usec / 1000000);

      if (debug > 1) {
        printf("DEBUG: updatedTime_sec = [%d]", (int)updatedTime_sec);
        printf(" updatedTime_usec = [%d]", (int)updatedTime_usec);
        printf(" carry_sec  = [%d]\n", (int)carry_sec);
        fflush(stdout);
      }

      updatedTime_usec -= carry_sec * 1000000;
      updatedTime_sec = nextTime_sec + offset_sec + carry_sec;

      sleepTime_sec = updatedTime_sec - now_sec;
      sleepTime_usec =
          (1000000 * sleepTime_sec + updatedTime_usec - now_usec) / playSpeed;

      if (debug > 1) {
        printf("DEBUG: updatedTime_sec = [%d]", (int)updatedTime_sec);
        printf(" updatedTime_usec = [%d]", (int)updatedTime_usec);
        printf(" --> sleepTime_sec = [%d]", (int)sleepTime_sec);
        printf(" sleepTime_usec = [%d]\n", (int)sleepTime_usec);
        fflush(stdout);
      }

      sleep_usec = sleepTime_usec;
      if (debug > 1) {
        printf("DEBUG: sleep_usec = [%d]\n", (int)sleepTime_usec);
        fflush(stdout);
      }
      if ((int)sleep_usec > 0) {
        hrtime_ew(&CurrentTime);
        if(stpoutput) {
          if( ScreenMsg && (CurrentTime - lastdot) > 1.0 )
          {
            lastdot = CurrentTime;
            fprintf(stdout,".");
            fflush(stdout);
          }
        }
        if (debug) {
          printf("DEBUG: Sleeping [%d] microseconds...\n", (int)sleep_usec);
          fflush(stdout);
        }
        usleep(sleep_usec);
      }
      //         }
      //         else
      //         {
      // drift = unow - prevTime - 1000000.*(secOffset);
      // drift = unow - nextTime - 1000000.*(secOffset);
      // nextTime -= drift;

      /*
      if(nextTime - prevTime > 0)
      {
              sleepTime = ((nextTime - prevTime)        / playSpeed);

              // sleep_ew((nextTime - prevTime) / 1000. * playSpeed)
              usleep(sleepTime);
      }
      */
      //        }
    }

    if (makeCurrent) {
      if (outputLogo.type == TypeTraceBuf2) // Wave records
      {
        // Update record timestamps
        trh->starttime += secOffset;
        trh->endtime += secOffset;

        // DEBUG: The following block is code for altering the message content
        // of a tracebuf record, and was used for intentionally replicating the
        // input from a particular "broken" station.  Could be useful for
        // creating simulations in future, if expanded upon.

        /*
        if(strcmp(trh->sta, "SNO") == 0)
        {
                int *data;
                data = (int *)(buffer + sizeof(TRACE2_HEADER));
                for(i=0; i<trh->nsamp; ++i)
                {
                        // printf("data[%d] = %d\n", i, data[i]);
                        data[i] = -1000;
                }
        }
        */

        /*
                                        if(strcmp(trh->sta, "BAI") == 0)
                                        {
                                                int *data;
                                                data = (int *)(buffer +
           sizeof(TRACE2_HEADER)); for(i=0; i<trh->nsamp; ++i)
                                                {
                                                        // printf("data[%d] =
           %d\n", i, data[i]); data[i] = 8388607;
                                                }
                                        }
        */

      } else if (outputLogo.type == TypePickSCNL) {
      } else if (outputLogo.type == TypeQuake2K) {
      } else if (outputLogo.type == TypeArc) {
      } else if (outputLogo.type == TypeCarlStaTrig) {
      } else if (outputLogo.type == TypeCarlSubTrig) {
      }
    }

    // Write output message to output ring
    if (activeTypes[(int)outputLogo.type] == 1) {
      if(stpoutput) {
        if(outputLogo.type == TypeTraceBuf2) // Wave records
        {
          gettimeofday(&tp, &tzp);
          if(firstout == 0) {
            double trueNow  = tp.tv_sec*1. + tp.tv_usec/1000000.;
            double trhStart = trh->starttime;
            double trhEnd   = trh->endtime;
            printf("DEBUG: >> FIRST_T_NOW = [%f]\n", trueNow);
            printf("DEBUG: >> FIRST_START = [%f]\n", trhStart);
            printf("DEBUG: >> FIRST_END   = [%f]\n", trhEnd);
            fflush(stdout);
            firstout = 1;
          }
        }
      }
      if (debug) {
        if (outputLogo.type == TypeTraceBuf2) // Wave records
        {
          gettimeofday(&tp, &tzp);
          double trueNow = tp.tv_sec * 1. + tp.tv_usec / 1000000.;
          double trhStart = trh->starttime;
          //    printf("t_now(s)=[%f] start_new(s)=[%f] diff(s)=[%f]
          //t_header(s)=[%f] start_orig=[%f] end_orig=[%f] lat_orig=[%f]
          //t_now-t_header=[%f]\n",                     trueNow, trhStart, (trueNow - trhStart),
          //nextTime/1000000., starttime_orig, endtime_orig,                    (nextTime/1000000.
          //- endtime_orig), (trueNow-nextTime/1000000.));
          printf("t_now(s)=[%f] start_new(s)=[%f] diff(s)=[%f] "
                 "t_header(s)=[%f] lat_orig=[%f]\n",
                 trueNow, trhStart, (trueNow - trhStart), nextTime / 1000000.,
                 (nextTime / 1000000. - endtime_orig));
          fflush(stdout);
        }
      }

      tport_putmsg(&outRegion, &outputLogo, (int)header.length, buffer);
    }

    // }
    /*
    else        // Non-wave records (Pick, Hypo, Trig, Subtrig, etc.)
    {
            if(fixedInterval)
            {
                    sleep_ew(interval*1000.);
            }
            else
            {
                    if(nextTime - prevTime > 0)
                    {
                            //sleep_ew((nextTime - prevTime) / (playSpeed));
                            // sleep_ew(1000.);
                    }
            }

    // Write output message to output ring
            if(activeTypes[(int) outputLogo.type] == 1)
            {
                    tport_putmsg(&outRegion, &outputLogo, (int) header.length,
    buffer);
            }

            // DEBUG
            if(debug > 1)
            {
                    printf("nextTime(us)=[%f] prevTime(us)=[%f]
    sleeptime(us)=[%f] --> \n", nextTime, prevTime,
    ((nextTime-prevTime)/playSpeed)); fflush(stdout);
            }
            // END DEBUG
    }
    */

    // Set reference times for next loop iteration
    prevTime = nextTime;

    prevTime_sec = nextTime_sec;
    prevTime_usec = nextTime_usec;
  }

  if (stpoutput) {
    printf("\n");
    fflush(stdout);
  }
  if (debug > 1) {
    printf("DEBUG: Out of main loop.\n");
    fflush(stdout);
  }

  if (!dump) {
    tport_detach(&outRegion);
  }
  close(fd); // Close input file

  return 0;
} // end main

// Function: lookup
void lookup(void) {
  // Specify logos to get

  if (GetType("TYPE_HEARTBEAT", &TypeHeartbeat) != 0) {
    fprintf(stderr, "%s: Error getting Type_Heartbeat.\n", MYNAME);
    fflush(stderr);
    exit(-1);
  }
  if (GetType((char *)"TYPE_ERROR", &TypeError) != 0) {
    fprintf(stderr, "%s: Invalid message type <TYPE_ERROR>; exiting!\n",
            MYNAME);
    exit(-1);
  }
  if (GetType("TYPE_TRACEBUF2", &TypeTraceBuf2) != 0) {
    fprintf(stderr, "%s: Invalid message type <TYPE_TRACEBUF2>; exiting!\n",
            MYNAME);
    fflush(stderr);
    exit(-1);
  }
  if (GetType((char *)"TYPE_PICK_SCNL", &TypePickSCNL) != 0) {
    fprintf(stderr, "%s: Invalid message type <TYPE_PICK_SCNL>; exiting!\n",
            MYNAME);
    fflush(stderr);
    exit(-1);
  }
  if (GetType((char *)"TYPE_QUAKE2K", &TypeQuake2K) != 0) {
    fprintf(stderr, "%s: Invalid message type <TYPE_QUAKE2K>; exiting!\n",
            MYNAME);
    fflush(stderr);
    exit(-1);
  }
  if (GetType((char *)"TYPE_HYP2000ARC", &TypeArc) != 0) {
    fprintf(stderr, "%s: Invalid message type <TYPE_HYP2000ARC>; exiting!\n",
            MYNAME);
    fflush(stderr);
    exit(-1);
  }
  if (GetType((char *)"TYPE_CARLSTATRIG", &TypeCarlStaTrig) != 0) {
    fprintf(stderr, "%s: Invalid message type <TYPE_CARLSTATRIG>; exiting!\n",
            MYNAME);
    fflush(stderr);
    exit(-1);
  }
  if (GetType((char *)"TYPE_TRIGLIST_SCNL", &TypeCarlSubTrig) != 0) {
    fprintf(stderr, "%s: Invalid message type <TYPE_TRIGLIST_SCNL>; exiting!\n",
            MYNAME);
    fflush(stderr);
    exit(-1);
  }
  //
  // Add additional clauses here for other types to register

} // end lookup

// Function: usage
void usage(void) {
  printf("\nUsage Syntax: \n");
  printf("-------------------------------------------------------------\n");
  printf("Usage 1:  fileplayer [configFile.d] [--stp] <infileName> <ringName>\n");
  printf("Usage 2:  fileplayer [configFile.d] [--stp] [-c] <infileName> <ringName> <startTime> <endTime>\n");
  printf("Usage 3:  fileplayer [configFile.d] [--stp] [-c] [-f <fixedInterval>] <infileName> <ringName> <startTime> <endTime>\n");
  printf("Usage 4:  fileplayer [configFile.d] [--stp] [-c] [-r <playbackRate>] <infileName> <ringName> <startTime> <endTime>\n");
  printf("Usage 5:  fileplayer [configFile.d] -d <infileName> <startTime> <endTime>\n");
  printf("\nRequired Arguments if not specified in configuration file:\n");
  printf("-------------------------------------------------------------\n");
  printf("<infileName> = Binary file with recorded EW ring data\n");
  printf("<ringName>   = EW ring to play data into\n");
  printf("\nOptional Arguments: \n");
  printf("-------------------------------------------------------------\n");
  printf("[configFile.d] = Configuration file name\n");
  printf("[--stp]        = include STP test output\n");
  printf("-d = dump recorded EW ring data\n");
  printf("[-c] = Update record timestamps to \"current\" time; simulate newly "
         "arrived data\n");
  printf("[-f] = Use fixed inter-message interval for playback (instead of "
         "original cadence)\n");
  printf("<fixedInterval> =  Forced inter-message interval (seconds) \n");
  printf("[-r] = Play back records with original cadence, at a user-defined "
         "speed\n");
  printf("<playbackRate>  =  Multiple to apply to original message cadence\n");
  printf(
      "<startTime> = Timestamp of beginning of playback window (Unix time)\n");
  printf("<endTime>   = Timestamp of end of playback window (Unix time)\n");
  printf("\nNotes:\n");
  printf("-------------------------------------------------------------\n");
  printf(">>> [-f] and [-r] are mutually exclusive.\n");
  printf(">>> If <startTime> is earlier than the earliest timestamp in the "
         "input file,\n");
  printf("    playback will start at the beginning of the file.\n");
  printf(">>> If <endTime> is later than the last timestamp in the input "
         "file,\n");
  printf("    playback will end at the end of the file.\n");
  printf("-------------------------------------------------------------\n");
  printf("Version: %s\n", VERSION);
} // end usage
