/**
    \file net.c
    \brief	Implementation of communication library which sets up a
                TCP connection between a client program and a server
 */

#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "net.h"
#define MAXSTR 4096

int smerr;

int net_debug = 0;

/* Utility routine to read a socket */
/* this version doesn't use select: */
static int read_buf(int sock, char *buf) {
  int i, n, nbuf, done, size;

  char rdbuf[MAXSTR];

  fd_set fdset;

  FD_ZERO(&fdset);
  FD_SET(sock, &fdset);

  size = sizeof(buf);

  bzero((char *)buf, size);
  nbuf = 0;
  done = 0;
  n = 0;

  while (!done) {

    /* read "rdbuf" */
    bzero((char *)rdbuf, sizeof(rdbuf));
    n = read(sock, rdbuf, sizeof(rdbuf));
    if (n <= 0) {
      smerr = SM_BROKENC;
      return (ERROR);
    }

    /* concatenate what we've read until we reach a string
     * terminator: */
    for (i = 0; i < n; ++i) {
      buf[nbuf] = rdbuf[i];

      if (rdbuf[i] == '\n' || rdbuf[i] == '\0') {
        buf[nbuf] = '\0';
        done = 1;
        break;
      }
      ++nbuf;
    }
  }
  return (nbuf);
}

int client_connect(char const *machine, int port) {
  struct hostent *host;
  struct protoent *proto;

  struct in_addr in;
  struct sockaddr_in sin;
  int sock, status;

  /* PROTOCOL */

  if ((proto = getprotobyname("tcp")) == 0) {
    smerr = SM_PROTO;
    return (ERROR);
  }

  /* SOCKET */

  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock < 0) {
    smerr = SM_SOCK;
    return (ERROR);
  }

  if (net_debug)
    fprintf(stderr, "client_connect: socket = %d\n", sock);

  /* FAMILY */

  bzero((char *)&sin, sizeof(struct sockaddr_in));
  sin.sin_family = AF_INET;

  /* PORT */

  if ((sin.sin_port = htons((u_short)port)) == 0) {
    smerr = SM_PORT;
    return (ERROR);
  }

  if (net_debug)
    fprintf(stderr, "client_connect: port = %d\n", port);

  /* HOST (name or IP address) */

  if ((host = gethostbyname(machine)))
    bcopy(host->h_addr, (char *)&sin.sin_addr, host->h_length);

  else if ((sin.sin_addr.s_addr = inet_addr(machine)) == INADDR_NONE) {
    smerr = SM_HOST;
    return (ERROR);
  }

  if (net_debug) {
    fprintf(stderr, "client_connect: host.h_name = %s\n", host->h_name);
    memcpy(&in.s_addr, host->h_addr_list[0], sizeof(in.s_addr));
    fprintf(stderr, "client_connect: host.h_addr = %s\n", inet_ntoa(in));
  }

  /* CONNECT */

  status = connect(sock, (struct sockaddr *)&sin, sizeof(sin));
  if (status < 0) {
    smerr = SM_CONNECT;
    return (ERROR);
  }

  if (net_debug)
    fprintf(stderr, "client_connect: connect OK\n");

  return (sock);
}

int client_send_cmd(int sock, char const *cmdin) {
  int len, offset;
  char cmd[MAXSTR];

  strcpy(cmd, cmdin);

  if (strlen(cmd) == 0) {
    smerr = SM_MESG;
    return (ERROR);
  }

  if (cmd[strlen(cmd) - 1] == '\n')
    cmd[strlen(cmd) - 1] = '\0';

  if (net_debug)
    fprintf(stderr, "client_send_cmd: \"%s\"\n", cmd);

  strcat(cmd, "\n");

  offset = 0;

  while (1) {
    len = write(sock, cmd + offset, strlen(cmd) - offset);

    if (len < 0) {
      smerr = SM_UNKNOWN;

      if (errno == EAGAIN)
        smerr = SM_EAGAIN;
      else if (errno == EWOULDBLOCK)
        smerr = SM_EWOULDBLOCK;
      else if (errno == EBADF)
        smerr = SM_EBADF;
      else if (errno == EDESTADDRREQ)
        smerr = SM_EDESTADDRREQ;
      else if (errno == EDQUOT)
        smerr = SM_EDQUOT;
      else if (errno == EFAULT)
        smerr = SM_EFAULT;
      else if (errno == EFBIG)
        smerr = SM_EFBIG;
      else if (errno == EINTR)
        smerr = SM_EINTR;
      else if (errno == EINVAL)
        smerr = SM_EINVAL;
      else if (errno == EIO)
        smerr = SM_EIO;
      else if (errno == ENOSPC)
        smerr = SM_ENOSPC;
      else if (errno == EPERM)
        smerr = SM_EPERM;
      else if (errno == EPIPE)
        smerr = SM_EPIPE;

      return (ERROR);
    }

    if (len == 0) {
      smerr = SM_ZERO;
      return (ERROR);
    }

    offset += len;

    if (offset == strlen(cmd))
      break;
  }

  return (OK);
}

int client_recv_cmd(int sock, char *string) {
  int n;

  char buf[MAXSTR];

  if ((n = read_buf(sock, buf)) < 0)
    return (ERROR);

  else {
    if (net_debug)
      fprintf(stderr, "client_recv_cmd: \"%s\"\n", buf);

    if (strncmp(buf, "ERROR: ", 7) == 0) {
      strcpy(string, "");
      smerr = atoi(buf + 10);

      if (net_debug) {
        fprintf(stderr, "client_recv_cmd: smerr = %d\n", smerr);
        fflush(stderr);
      }

      return (ERROR);
    } else {
      strcpy(string, buf);
      return (OK);
    }
  }
}

int server_connect(int port) {
  struct protoent *proto;
  struct hostent *host;
  struct in_addr in;
  struct sockaddr_in sin, new_sin;
  int status, sock, new_sock;
  socklen_t new_len;
  /*
     struct timeval      tp1, tp2;
     struct timezone     tzp;
  */

  char hostname[MAXSTR];

  (void)signal(SIGCHLD, SIG_IGN);

  /* PROTOCOL */

  proto = getprotobyname("tcp");
  if (proto == 0) {
    smerr = SM_PROTO;
    return (ERROR);
  }

  /* SOCKET */

  sock = socket(AF_INET, SOCK_STREAM, proto->p_proto);
  if (sock < 0) {
    smerr = SM_SOCK;
    return (ERROR);
  }

  if (net_debug)
    fprintf(stderr, "server_connect: socket = %d\n", sock);

  /* FAMILY (AF_INET = 2) */

  bzero((char *)&sin, sizeof(struct sockaddr_in));
  sin.sin_family = AF_INET;

  /* PORT */

  if ((sin.sin_port = htons((u_short)port)) == 0) {
    smerr = SM_PORT;
    return (ERROR);
  }

  if (net_debug)
    fprintf(stderr, "server_connect: port = %d\n", port);

  /* HOST */

  gethostname(hostname, MAXSTR);
  host = gethostbyname(hostname);
  bcopy(host->h_addr, (char *)&sin.sin_addr, host->h_length);

  if (net_debug) {
    fprintf(stderr, "server_connect: host.h_name = %s\n", host->h_name);
    memcpy(&in.s_addr, host->h_addr_list[0], sizeof(in.s_addr));
    fprintf(stderr, "server_connect: host.h_addr = %s\n", inet_ntoa(in));
  }

  /* BINDING */

  status = bind(sock, (struct sockaddr *)&sin, sizeof(sin));
  if (status < 0) {
    if (net_debug) {
      fprintf(stderr, "server_connect: bind() failed, condition %d\n", errno);
      fflush(stderr);
    }

    smerr = SM_BIND;
    return (ERROR);
  }

  if (net_debug)
    fprintf(stderr, "server_connect: bind() OK\n");

  /* LISTEN and ACCEPT Connection */

  status = listen(sock, 5);
  if (status < 0) {
    smerr = SM_LISTEN;
    return (ERROR);
  }

  if (net_debug)
    fprintf(stderr, "server_connect: listen() OK\n");

  new_len = sizeof(new_sin);
  while (1) {
    new_sock = accept(sock, (struct sockaddr *)&new_sin, &new_len);
    if (new_sock < 0) {
      smerr = SM_ACCEPT;
      return (ERROR);
    }

    if (net_debug)
      fprintf(stderr, "server_connect: accept() new_sock = %d\n", new_sock);
    /*
        gettimeofday(&tp1, &tzp);
        printf("start time tp1: %ld sec. %ld msec.\n", tp1.tv_sec,
            tp1.tv_usec);
            fork();
        gettimeofday(&tp2, &tzp);
        printf("end time tp2: %ld sec. %ld msec.\n", tp2.tv_sec,
            tp2.tv_usec);
    */

    if (fork() == 0) /* Child */
    {
      close(sock);
      return (new_sock);
    } else /* Parent */
      close(new_sock);
  }
}

int server_send_cmd(int sock, char const *cmdin) {
  int len;
  char cmd[MAXSTR];

  strcpy(cmd, cmdin);

  if (cmd[strlen(cmd) - 1] == '\n')
    cmd[strlen(cmd) - 1] = '\0';

  if (net_debug)
    fprintf(stderr, "server_send_cmd: \"%s\"\n", cmd);

  strcat(cmd, "\n");

  len = write(sock, cmd, strlen(cmd));

  if (len < (int)strlen(cmd))
    return (ERROR);
  else
    return (OK);
}

int server_recv_cmd(int sock, char *string) {
  int n;
  char buf[MAXSTR];

  if ((n = read_buf(sock, buf)) < 0)
    return (ERROR);

  else {
    if (net_debug)
      fprintf(stderr, "server_recv_cmd: \"%s\"\n", buf);

    strcpy(string, buf);
    return (OK);
  }
}
