/*
 * strfuncs.c
 *
 *  string functions
 */
#include "strfuncs.h"
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>



char *strndup_loc(char *str, int chars)
{
    char *buffer;
    int n;

    buffer = (char *) malloc(chars +1);
    if (buffer)
    {
        for (n = 0; ((n < chars) && (str[n] != 0)) ; n++) buffer[n] = str[n];
        buffer[n] = 0;
    }

    return buffer;
}


/**
 * Free the argument vector with last entry being a NULL terminator.
 * @param argv the argument vector;
 */
void freeargv(char *argv[]) {
  int argc;
  for (argc = 0; argv[argc]; argc++) {
    free(argv[argc]);
    argv[argc] = NULL;
  }
  free(argv);
}

/**
 * Copy string src to buffer dst of size n.  At most n-1 non-null chars
 * will be copied.  Always null terminates (unless n == 0) the buffer.
 * This should replace strncpy which does not always null terminate.
 */
char *strcpyn(char *dst, const char *src, size_t n)
{
  if (n)
  {
    const size_t srclen = strlenn(src, n);
    if (srclen + 1 < n)
    {
      memcpy(dst, src, srclen + 1);
    }
    else
    {
      memcpy(dst, src, n - 1);
      dst[n - 1] = '\0';
    }
  }
  return (dst);
}

/**
 * strendswith:
 * Tests if the  specified string ends with the specified suffix.
 * @str the string
 * @suffix the suffix
 * @return non-zero  if the  specified string ends with the specified suffix;
 * zero otherwise
 */
int strendswith(char *const str, char *const suffix) {
  size_t slen = strlen(str);
  size_t suffixlen = strlen(suffix);
  if (suffixlen > slen)
    return 0;
  return strcmp(str + slen - suffixlen, suffix) == 0;
}

/*
 * The strlenn() function computes the length of the string str,
 * but stops when the length reaches the maximum of n.
 */
size_t strlenn(const char *str, size_t n)
{
  int len = 0;
  const char *s;
  for (s = str; *s; ++s)
  {
     if (++len >= n)
     {
       break;
     }
  }
  return len;
}

/**
 * strreplace:
 * @str: a string to replace characters in
 * @strsize: the size of the str array
 * @oldChar: the character to be replaced
 * @newChar: the character to replace
 *
 * Modifies all old characters from a string with the specified
 * new character.
 *
 * This function doesn't allocate or reallocate any memory;
 * it modifies @str in place. Therefore, it cannot be used on
 * statically allocated strings.
 *
 * The pointer to @str is returned to allow the nesting of functions.
 *
 * Returns: @str
 */
char *strreplace(char *str, size_t strsize, char oldChar, char newChar) {
  int i;
  for (i = 0; i < strsize; i++) {
    if (str[i] == oldChar) {
      str[i] = newChar;
    }
  }
  return str;
}

/**
 * strreplace_string:
 * @orig: a string to replace characters in
 * @rep:  substring to be replaced
 * @with: new substring
 *
 * Replace substring in the string
 *
 * !!!!!!!!!You must free the result if result is non-NULL.!!!!!!!!!!!!!!!!
 *
 * The pointer to @str is returned to allow the nesting of functions.
 *
 * Returns: new allocated modified string
 */


char *strreplace_string(char *orig, char *rep, char *with) {
    char *result; // the return string
    char *ins;    // the next insert point
    char *tmp;    // varies
    int len_rep;  // length of rep (the string to remove)
    int len_with; // length of with (the string to replace rep with)
    int len_front; // distance between rep and end of last rep
    int count;    // number of replacements

    // sanity checks and initialization
    if (!orig || !rep)
        return NULL;
    len_rep = strlen(rep);
    if (len_rep == 0)
        return NULL; // empty rep causes infinite loop during count
    if (!with)
        with = "";
    len_with = strlen(with);

    // count the number of replacements needed
    ins = orig;
    for (count = 0; tmp = strstr(ins, rep); ++count) {
        ins = tmp + len_rep;
    }

    tmp = result = malloc(strlen(orig) + (len_with - len_rep) * count + 1);

    if (!result)
        return NULL;

    // first time through the loop, all the variable are set correctly
    // from here on,
    //    tmp points to the end of the result string
    //    ins points to the next occurrence of rep in orig
    //    orig points to the remainder of orig after "end of rep"
    while (count--) {
        ins = strstr(orig, rep);
        len_front = ins - orig;
        tmp = strncpy(tmp, orig, len_front) + len_front;
        tmp = strcpy(tmp, with) + len_with;
        orig += len_front + len_rep; // move to next "end of rep"
    }
    strcpy(tmp, orig);
    return result;
}

/**
 * strstartswith:
 * Tests if the  specified string starts with the specified prefix.
 * @str the string
 * @prefix the suffix
 * @return non-zero  if the  specified string starts with the specified prefix;
 * zero otherwise
 */
int strstartswith(char *const str, char *const prefix) {
  return strncmp(str, prefix, strlen(prefix)) == 0;
}

/**
 * strstrip:
 * @str: a string to remove the leading and trailing whitespace from
 *
 * Removes leading and trailing whitespace from a string, by moving the rest
 * of the characters forward.
 *
 * This function doesn't allocate or reallocate any memory;
 * it modifies @str in place. Therefore, it cannot be used on
 * statically allocated strings.
 *
 * The pointer to @str is returned to allow the nesting of functions.
 *
 * Returns: @str
 */
char *strstrip(char *const str) {
  strstripLeading(str);
  strstripTrailing(str);
  return str;
}

/**
 * strstripLeading:
 * @str: a string to remove the leading whitespace from
 *
 * Removes leading whitespace from a string, by moving the rest
 * of the characters forward.
 *
 * This function doesn't allocate or reallocate any memory;
 * it modifies @str in place. Therefore, it cannot be used on
 * statically allocated strings.
 *
 * The pointer to @str is returned to allow the nesting of functions.
 *
 * Returns: @str
 */
char *strstripLeading(char *const str) {
  if (str != NULL) {
    unsigned char *ustr;
    for (ustr = (unsigned char *)str; *ustr && isspace(*ustr); ustr++)
      ;
    if ((unsigned char *)str != ustr) {
      memmove(str, ustr, strlen((char *)ustr) + 1);
    }
  }
  return str;
}

/**
 * strstripTrailing:
 * @str: a string to remove the trailing whitespace from
 *
 * Removes trailing whitespace from a string.
 *
 * This function doesn't allocate or reallocate any memory;
 * it modifies @str in place. Therefore, it cannot be used
 * on statically allocated strings.
 *
 * The pointer to @str is returned to allow the nesting of functions.
 *
 * Returns: @str
 */
char *strstripTrailing(char *const str) {
  if (str != NULL) {
    int len = strlen(str);
    while (len--) {
      if (!isspace((unsigned char)str[len])) {
        break;
      }
      str[len] = '\0';
    }
  }
  return str;
}

/**
 * Convert string to argument vector with last entry being a NULL terminator.
 * The memory obtained is done dynamically using calloc and hence it can be
 * freed using freeargv().
 * @param str the string.
 * @param delim the deliimter character, typically the space character.
 * @return the argument vector.
 */
char **str2argv(char *const str, const char delim) {
  int argc;
  char **argv;
  char *strp;
  int charcount;
  // determine the number of strings
  int strcount = 1;
  strp = str;
  while (*strp != '\0') {
    if (*strp++ == delim) {
      strcount++;
    }
  }
  // allocate array with extra entry for NULL at end
  argv = calloc((strcount + 1), sizeof(char *));
  if (argv == NULL) {
    return NULL;
  }

  strp = str;
  charcount = 0;
  for (argc = 0;;) {
    if (*strp == delim || *strp == '\0') {
      argv[argc] = strndup_loc(strp - charcount, charcount);
      if (argv[argc] == NULL) { // if memory allocation failed
        freeargv(argv);         // free previous memory
        return NULL;
      }
      if (*strp == '\0') {
        break;
      }
      charcount = 0;
      argc++;
    } else {
      charcount++;
    }
    strp++;
  }
  return argv;
}

